/************************************************************************/
/*  TP GPU - GLSL Ensimag 3A                                            */
/*  Sbastien Barbier - EVASION                                         */
/*	janvier 2009														*/
/************************************************************************/

#include "headers.h"
#include "structure.h"
#include "globalOptions.h"
#include "trackball.h"
#include "IO.h"
#include "shadersgeo.h"


//GLU variable

static int main_window;
static int window_height = 512;
static int window_width  = 512;

static global gbl;
static int beginx, beginy;

static int n_vertex;
static int n_face;

static Vertex *list_v;
static Face   *list_f;

static GLhandleARB shader_silhouette;

int adjacentVertex(int face, int neighbor)
{
	int adjacent_face = list_f[face].F[neighbor];

	for(int v=0; v<3; v++){

		if (list_f[adjacent_face].V[v] != list_f[face].V[(neighbor+1)%3] &&
			list_f[adjacent_face].V[v] != list_f[face].V[(neighbor+2)%3])
			return list_f[adjacent_face].V[v];
	}

	fprintf(stderr, "Grosse erreur dans le fichier .NGB\n");
	exit(0);

}
void drawBoundingBox()
{
  glLineWidth(2);
  
  glColor3f(0.,0.,1.);
  glBegin(GL_LINE_STRIP);
  glVertex3d( gbl.BB[0], gbl.BB[1], gbl.BB[2]);
  glVertex3d( gbl.BB[3], gbl.BB[1], gbl.BB[2]);
  glVertex3d( gbl.BB[3], gbl.BB[4], gbl.BB[2]);
  glVertex3d( gbl.BB[0], gbl.BB[4], gbl.BB[2]);
  glVertex3d( gbl.BB[0], gbl.BB[1], gbl.BB[2]);
  glEnd();
  
  glBegin(GL_LINE_STRIP);
  glVertex3d( gbl.BB[0], gbl.BB[1], gbl.BB[5]);
  glVertex3d( gbl.BB[3], gbl.BB[1], gbl.BB[5]);
  glVertex3d( gbl.BB[3], gbl.BB[4], gbl.BB[5]);
  glVertex3d( gbl.BB[0], gbl.BB[4], gbl.BB[5]);
  glVertex3d( gbl.BB[0], gbl.BB[1], gbl.BB[5]);
  glEnd();

  glBegin(GL_LINES);
  glVertex3d( gbl.BB[0], gbl.BB[1], gbl.BB[2]);
  glVertex3d( gbl.BB[0], gbl.BB[1], gbl.BB[5]);
  glVertex3d( gbl.BB[3], gbl.BB[1], gbl.BB[2]);
  glVertex3d( gbl.BB[3], gbl.BB[1], gbl.BB[5]);
  glVertex3d( gbl.BB[0], gbl.BB[4], gbl.BB[2]);
  glVertex3d( gbl.BB[0], gbl.BB[4], gbl.BB[5]);
  glVertex3d( gbl.BB[3], gbl.BB[4], gbl.BB[2]);
  glVertex3d( gbl.BB[3], gbl.BB[4], gbl.BB[5]);
  glEnd();

  glLineWidth(1);
} 

void drawVertices()
{

	glColor3f(0,1,0);
	glBegin(GL_POINTS);
	for(int i=0;i<n_vertex;i++){
		glVertex3fv(list_v[i].V);
	}
	glEnd();
}

// compute normal on the-fly 
// it is not optimal !
void computeNormal(float v0[3], float v1[3], float v2[3], float res[3])
{
	float n1[3], n2[3];

	n1[0] = v1[0] - v0[0];
	n2[0] = v2[0] - v0[0];
	n1[1] = v1[1] - v0[1];
	n2[1] = v2[1] - v0[1];
	n1[2] = v1[2] - v0[2];
	n2[2] = v2[2] - v0[2];

	res[0] =  n1[1] * n2[2] - n1[2] * n2[1];
	res[1] = -n1[0] * n2[2] + n1[2] * n2[0];
	res[2] =  n1[0] * n2[1] - n1[1] * n2[0];
	
	float norm = sqrt(res[0]*res[0]+res[1]*res[1]+res[2]*res[2]);

	res[0] /= norm;
	res[1] /= norm;
	res[2] /= norm;

}

void drawEdges()
{

	glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LESS); 

	glColor3f(0.5,0.5,0.5);
	glBegin(GL_LINES);
	for(int i=0;i<n_face;i++)
	{
		glVertex3fv(list_v[list_f[i].V[0]].V);
		glVertex3fv(list_v[list_f[i].V[1]].V);
		glVertex3fv(list_v[list_f[i].V[1]].V);
		glVertex3fv(list_v[list_f[i].V[2]].V);
		glVertex3fv(list_v[list_f[i].V[2]].V);
		glVertex3fv(list_v[list_f[i].V[0]].V);

	}
	glEnd();
	glDisable(GL_LIGHTING);
	glEnable(GL_COLOR_MATERIAL); 

	glDisable(GL_DEPTH_TEST);
}

void drawFaces()
{

	float normal[3];

	glShadeModel(GL_SMOOTH);
	glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LESS); 

	glColor3f(1,1,1);
	glEnable(GL_COLOR_MATERIAL);
	glEnable(GL_LIGHTING);
	glBegin(GL_TRIANGLES);
	for(int i=0;i<n_face;i++)
	{
		computeNormal(list_v[list_f[i].V[0]].V, 
					  list_v[list_f[i].V[1]].V, 
					  list_v[list_f[i].V[2]].V, 
					  normal);
		glNormal3fv(normal);
		glVertex3fv(list_v[list_f[i].V[0]].V);
		glVertex3fv(list_v[list_f[i].V[1]].V);
		glVertex3fv(list_v[list_f[i].V[2]].V);
	}
	glEnd();
	glDisable(GL_LIGHTING);
	glEnable(GL_COLOR_MATERIAL);
}

void Display(void)
{
	
	int next_v;

	glViewport(0, 0, window_width, window_height);

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);  
	glClearColor(0,0,0,0);
	glMatrixMode( GL_PROJECTION ) ;
	glLoadIdentity() ;

	glOrtho (gbl.ortho[0] + gbl.maxRange/(float)100,
	         gbl.ortho[1] - gbl.maxRange/(float)100,
	         gbl.ortho[2] + gbl.maxRange/(float)100,
		 gbl.ortho[3] - gbl.maxRange/(float)100,
		 gbl.ortho[4], 
		 gbl.ortho[5]);

	
	// CAMERA
	gluLookAt(0., 0., 2*gbl.maxRange, 0., 0., 0., 0., 1., 0.);

	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();
	build_rotmatrix(gbl.M_modelView, gbl.current_quat, gbl.translation);
	glMultMatrixf(&gbl.M_modelView[0][0]);

	glTranslatef( -gbl.origin[0], -gbl.origin[1], -gbl.origin[2]); 

	if (gbl.bounding_box) drawBoundingBox();
	if (gbl.vertices)     drawVertices();
	if (gbl.edges)		  drawEdges();
	if (gbl.faces)		  drawFaces();


	
	glUseProgramObjectARB(shader_silhouette);

	glBegin(GL_TRIANGLES_ADJACENCY_EXT);

	for(int f=0; f<n_face;f++){
		
		glVertex3fv(list_v[list_f[f].V[0]].V);
		next_v = adjacentVertex(f,2);
		glVertex3fv(list_v[next_v].V);
		glVertex3fv(list_v[list_f[f].V[1]].V);
		next_v = adjacentVertex(f,0);
		glVertex3fv(list_v[next_v].V);
		glVertex3fv(list_v[list_f[f].V[2]].V);
		next_v = adjacentVertex(f,1);
		glVertex3fv(list_v[next_v].V);

	}
	glEnd();

	glUseProgramObjectARB(0);

	glutSwapBuffers();
	
}

void Reshape(int w, int h) 
{  

	window_width  = w;
	window_height = h;
    glViewport(0,0, window_width, window_height);
	glutPostRedisplay();
}

void Mouse ( int button, int state, int x, int y )
{ 
	beginx = x;
	beginy = y;

	switch(button)
	{
	case GLUT_LEFT_BUTTON:
		if( state == GLUT_DOWN )  gbl.action = ROTATE; 
		else                      gbl.action = ROTATE_END; 
		break ;

	case GLUT_MIDDLE_BUTTON:
		if( state == GLUT_DOWN )  gbl.action = ZOOM;
		else                      gbl.action = ZOOM_END;
		break ;

	case GLUT_RIGHT_BUTTON:
		if( state == GLUT_DOWN )  gbl.action = TRANSLATE;
		else                      gbl.action = TRANSLATE_END;	
		break ;	
	}
}



void Motion ( int x, int y )
{

	switch(gbl.action)
	{
	case ROTATE:
		trackball(gbl.last_quat,
			(2.0 * beginx - window_width) / (double) (2*window_width),
			(window_height - 2.0 * beginy) /(double) (2*window_height),
			(2.0 * x - window_width) / (double) (2*window_width),
			(window_height - 2.0 * y) / (double) (2*window_height));
		beginx = x;
		beginy = y;
		add_quats(gbl.last_quat, gbl.current_quat, gbl.current_quat);
		glutPostRedisplay();
		break ;

	case TRANSLATE:  
		gbl.translation[0] += (x-beginx)*gbl.step;
		gbl.translation[1] += (beginy-y)*gbl.step;
		beginx = x;
		beginy = y;
		glutPostRedisplay();
		break ;

	case ZOOM:
		if ((gbl.ortho[0] + (y - beginy)*gbl.step) < -0.005) {
			gbl.ortho[0] +=(y-beginy)*gbl.step;
			gbl.ortho[1] -=(y-beginy)*gbl.step;
			gbl.ortho[2] +=(y-beginy)*gbl.step;
			gbl.ortho[3] -=(y-beginy)*gbl.step;
		}
		beginy = y;
		glutPostRedisplay();
		break;
	default:
		break;
	}
}




void displayHelp()
{
	cout << "HELP\n"
		 << "B: display bounding-box\n"
		 << "F: display faces\n"
		 << "V: display vertices\n"
		 << "W: display edges\n"
		 << "H: display help\n"
		 << "Q: exit program\n"
		 << "MOUSE LEFT: rotate\n"
		 << "MOUSE RIGHT: translate\n"
		 << "MOUSE MIDDLE: zoom"
		 << endl;

}
void Keyboard(unsigned char key, int x, int y)
{ 	
	switch (key)
	{ 
	case 'b':
		gbl.bounding_box = (gbl.bounding_box+1) %2;
		glutPostRedisplay();
		break;
	case 'f':
		gbl.faces = (gbl.faces+1) %2;
		glutPostRedisplay();
		break;
	case 'v':
		gbl.vertices = (gbl.vertices+1) %2;
		glutPostRedisplay();
		break;
	case 'w':
		gbl.edges = (gbl.edges+1) %2;
		glutPostRedisplay();
		break;
	case 'h':
		displayHelp();
		glutPostRedisplay();
		break;
	case 'q':
		exit(0);
		break;
	default:
		break;
	}

}  



void initView()
{
	gbl.maxRange = fabs(gbl.BB[3]-gbl.BB[0]);
	if (fabs(gbl.BB[4]-gbl.BB[1]) > gbl.maxRange)
		gbl.maxRange = fabs(gbl.BB[4]-gbl.BB[1]);
	if (fabs(gbl.BB[5]-gbl.BB[2]) > gbl.maxRange)
		gbl.maxRange = fabs(gbl.BB[5]-gbl.BB[2]);
	
	gbl.step     = gbl.maxRange/100.;

	gbl.origin[0] = fabs(gbl.BB[3]-gbl.BB[0])/2.+gbl.BB[0];
	gbl.origin[1] = fabs(gbl.BB[4]-gbl.BB[1])/2.+gbl.BB[1];
	gbl.origin[2] = fabs(gbl.BB[5]-gbl.BB[2])/2.+gbl.BB[2];

	gbl.translation[0] = 0.;
	gbl.translation[1] = 0.;
	gbl.translation[2] = 0.;

	gbl.ortho[0] =    -gbl.maxRange * 2*1.5;  
	gbl.ortho[1] =    gbl.maxRange * 2*1.5;  
	gbl.ortho[2] =	  -gbl.maxRange * 2*1.5;  
	gbl.ortho[3] =    gbl.maxRange * 2*1.5;  
	gbl.ortho[4] =    gbl.step/100;
	gbl.ortho[5] =    4*gbl.maxRange;

	trackball(gbl.current_quat, 0.0, 0.0, 0.0, 0.0);

	gbl.bounding_box  = 0;
	gbl.vertices      = 0;
	gbl.faces         = 0;
	gbl.edges		  = 0;


	// LIGHTS, MATERIALS
	GLfloat direction[4];
	direction[0]=0.0; 
	direction[1]=0.0; 
	direction[2]=0.0;
	direction[3]=1.0; 
	glLightfv(GL_LIGHT0, GL_POSITION, direction);
	float MatSpec[4] = {1.0f, 1.0f, 1.0f, 1.0f};
	float MatDif[4] = {1.0f, 1.0f, 1.0f, 1.0f};
	float MatAmb[4] = {0.3f, 0.3f, 0.3f, 1.0f};
	float Light1Dif[4] = {1.0f, 0.2f, 0.2f, 1.0f};
	float Light1Spec[4] = {1.0f, 0.2f, 0.2f, 1.0f};
	float Light1Amb[4] = {0.5f, 0.5f, 0.5f, 1.0f};
	glMaterialfv(GL_FRONT_AND_BACK,GL_SPECULAR,MatSpec); 	//On applique les paramtres du matriau
	glMaterialfv(GL_FRONT_AND_BACK,GL_DIFFUSE,MatDif);
	glMaterialfv(GL_FRONT_AND_BACK,GL_AMBIENT,MatAmb);
	glMateriali(GL_FRONT_AND_BACK,GL_SHININESS,100); 
	glLightfv(GL_LIGHT0, GL_DIFFUSE, Light1Dif); 	//Et ceux de la lumire
	glLightfv(GL_LIGHT0, GL_SPECULAR, Light1Spec);
	glLightfv(GL_LIGHT0, GL_AMBIENT, Light1Amb);

	glEnable(GL_LIGHT0);
	
}

int main( int argc, char *argv[])
{ 
	
	FILE* fin;
	char s_file_ngb[100];

	if (argc > 2){
		fprintf(stderr, "%s file.ngb\n", argv[0]);
		exit(-1);
	}

	if (argc == 1)
		memcpy(s_file_ngb, "horse.ngb", sizeof("horse.ngb"));
	else
		memcpy(s_file_ngb, argv[1], sizeof(argv[1]));
	
	// -----------------------------------------------
	// INTIALIZE THE WINDOW 
	// -----------------------------------------------
	glutInit(&argc, argv);

	// -----------------------------------------------
	// INITIALIZE THE GRAPHIC RENDERING WINDOW 
	// -----------------------------------------------
	glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH);

	glutInitWindowSize(window_width , window_height);
	glutInitWindowPosition(10,10);
	main_window = glutCreateWindow("Silhouette");

	glDisable(GL_LIGHTING);

	GLenum err = glewInit();
	if (GLEW_OK != err)
	{
		/* Problem: glewInit failed, something is seriously wrong. */
		fprintf(stderr, "Error: %s\n", glewGetErrorString(err));
  
	}
    fprintf(stdout, "Status: Using GLEW %s\n", glewGetString(GLEW_VERSION));

	if((fin = fopen(s_file_ngb, "r")) == NULL){
		fprintf(stderr,  "Cannot open file %s\n", s_file_ngb);
		exit(1);
	}

	ReadNeighbourFile(fin, &n_vertex, &n_face, &list_v, &list_f, 
						   &gbl.BB[0], &gbl.BB[3], &gbl.BB[1], &gbl.BB[4], &gbl.BB[2], &gbl.BB[5]);

	fclose(fin);

	initShadersGeo(&shader_silhouette, "shaders/silhouette");
	initView();

	displayHelp();
	
	glutMouseFunc( Mouse );
	glutMotionFunc( Motion );
	glutKeyboardFunc( Keyboard );
	glutReshapeFunc( Reshape );
	glutDisplayFunc( Display );

	glutMainLoop();

	return 1;
}

